"""

"""
from __future__ import absolute_import
from __future__ import print_function
import csv
from datetime import datetime
from .._common import BaseServer


########################################################################
class LogManager(BaseServer):
    """
    Helper class for the management of logs by administrators. 
    
    Logs are the transaction records written by the various components 
    of ArcGIS Server.  You can query the logs, change various log settings, 
    and check error messages for helping to determine the nature of an issue.

    """
    _url = None
    _con = None
    _json_dict = None
    _json = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis,
                 initialize=False):
        """Constructor

               
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        url                    Required string. The machine URL.
        ------------------     --------------------------------------------------------------------
        gis                    Optional string. The GIS or Server object..
        ==================     ====================================================================
        
        """
        connection = gis
        super(LogManager, self).__init__(gis=gis,
                                  url=url)
        self._url = url
        self._con = connection
        if initialize:
            self._init(connection)
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def count_error_reports(self, machine="*"):
        """
        This operation counts the number of error reports (crash reports) that have been generated 
        on each machine.
            
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        machine                Optional string. The name of the machine on which to count the 
                               reports. The default will return the count for all machines in a site.
        ==================     ====================================================================

        :return:
           A dictionary with machine name and report count.
        
        """
        params = {
            "f": "json",
            "machine" : machine
        }
        url = self._url + "/countErrorReports"
        return self._con.post(path=url,
                              postdata=params)
    #----------------------------------------------------------------------
    def clean(self):
        """
        Deletes all the log files on all server machines in the site. This is an irreversible
        operation.
        
        This operation forces the server to clean the logs, which has the effect of freeing 
        up disk space. However, it is not required that you invoke this operation because 
        the server periodically purges old logs.
        
        :return:
           A boolean indicating success (True) or failure (False).
           
        """
        params = {
            "f" : "json",
        }
        url = "{}/clean".format(self._url)
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    @property
    def settings(self):
        """Gets the current log settings. """
        params = {
            "f" : "json"
        }
        url = self._url + "/settings"
        try:
            return self._con.post(path=url,
                                  postdata=params)['settings']
        except:
            return ""
    #----------------------------------------------------------------------
    def edit(self,
             level="WARNING",
             log_dir=None,
             max_age=90,
             max_report_count=10):
        """
        Provides log editing capabilities for the entire site. 
           
        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        level                  Optional string. The log level.  Can be one of (in severity order): 
                               OFF, DEBUG, VERBOSE, FINE, INFO, WARNING, SEVERE. The default is WARNING.
        ------------------     --------------------------------------------------------------------
        log_dir                Optional string. The file path to the root of the log directory.
        ------------------     --------------------------------------------------------------------
        max_age                Optional integer. The number of days that a server should save a 
                               log file. The default is 90.
        ------------------     --------------------------------------------------------------------
        ax_report_count        Optional integer. The maximum number of error report files per 
                               machine. The default is 10.
        ==================     ====================================================================
        

        :return:
           A JSON with the edited settings.
           
        """
        url = self._url + "/settings/edit"
        allowed_levels = ("OFF", "SEVERE", "WARNING", "INFO", "FINE", "VERBOSE", "DEBUG")
        current_settings = self.settings
        current_settings["f"] = "json"

        if level.upper() in allowed_levels:
            current_settings['logLevel'] = level.upper()
        if log_dir is not None:
            current_settings['logDir'] = log_dir
        if max_age is not None and \
           isinstance(max_age, int):
            current_settings['maxLogFileAge'] = max_age
        if max_report_count is not None and \
           isinstance(max_report_count, int) and\
           max_report_count > 0:
            current_settings['maxErrorReportsCount'] = max_report_count
        return self._con.post(path=url,
                              postdata=current_settings)
    #----------------------------------------------------------------------
    def query(self,
              start_time=None,
              end_time=None,
              since_server_start=False,
              level="WARNING",
              services="*",
              machines="*",
              server="*",
              codes=None,
              process_IDs=None,
              export=False,
              export_type="CSV", #CSV or TAB
              out_path=None):
        """
        The query operation on the logs resource provides a way to
        aggregate, filter, and page through logs across the entire site.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        start_time             Optional string. The most recent time to query.  Default is now. 
                               Time can be specified in milliseconds since UNIX epoch, or as an 
                               ArcGIS Server timestamp. For example { "startTime": "2011-08-01T15:17:20,123", ... },
                               { "startTime": 1312237040123, ... }, respectively.
        ------------------     --------------------------------------------------------------------
        end_time               Optional string. The oldest time to include in the result set. You 
                               can use this to limit the query to the last n minutes or hours as 
                               needed. Default is the beginning of all logging.
        ------------------     --------------------------------------------------------------------
        since_server_start     Optional string. Gets only the records written since the server 
                               started (True).  The default is False.
        ------------------     --------------------------------------------------------------------
        level                  Optional string. Gets only the records with a log level at or more 
                               severe than the level declared here. Can be one of (in severity 
                               order): DEBUG, VERBOSE, FINE, INFO, WARNING, SEVERE. The 
                               default is WARNING.
        ------------------     --------------------------------------------------------------------
        services               Optional string. Query records related to a specific service. 
                               The default is all.
        ------------------     --------------------------------------------------------------------
        machines               Optional string. Query records related to a specific machine. 
                               The default is all.
        ------------------     --------------------------------------------------------------------
        server                 Optional string. Query records related to a specific server. 
                               The default is all.
        ------------------     --------------------------------------------------------------------
        codes                  Optional string. Gets only the records with the specified code.  
                               The default is all.  See http://server.arcgis.com/en/server/latest/administer/windows/log-codes-overview.htm
        ------------------     --------------------------------------------------------------------
        process_IDs            Optional string. Query by the machine process ID that logged the event.
        ------------------     --------------------------------------------------------------------
        export                 Optional string. Boolean indicating whether to export the query 
                               results.  The default is False (don't export).
        ------------------     --------------------------------------------------------------------
        export_type            Optional string. The export file type. CSV or TAB are the choices, 
                               CSV is the default.
        ------------------     --------------------------------------------------------------------
        out_path               Optional string. The path to download the log file to.
        ==================     ====================================================================
        
        :return:
           A JSON of the log items that match the query. If export option is set to True, the 
           output log file path is returned.
        
        
        """
     
    
        if codes is None:
            codes = []
        if process_IDs is None:
            process_IDs = []
        allowed_levels = ("SEVERE", "WARNING", "INFO",
                          "FINE", "VERBOSE", "DEBUG")
        qFilter = {
            "services": "*",
            "machines": "*",
            "server" : "*"
        }
        if len(process_IDs) > 0:
            qFilter['processIds'] = process_IDs
        if len(codes) > 0:
            qFilter['codes'] = codes
        params = {
            "f" : "json",
            "sinceServerStart" : since_server_start,
            "pageSize" : 10000
        }
        url = "{url}/query".format(url=self._url)
        if start_time is not None and \
           isinstance(start_time, datetime):
            params['startTime'] = start_time.strftime("%Y-%m-%dT%H:%M:%S")
        if end_time is not None and \
           isinstance(end_time, datetime):
            params['endTime'] = end_time.strftime("%Y-%m-%dT%H:%M:%S")
        if level.upper() in allowed_levels:
            params['level'] = level
        if server != "*":
            qFilter['server'] = server.split(',')
        if services != "*":
            qFilter['services'] = services.split(',')
        if machines != "*":
            qFilter['machines'] = machines.split(",")
        params['filter'] = qFilter
        if export is True and \
           out_path is not None:

            messages = self._con.post(path=url,
                                      postdata=params)
            with open(name=out_path, mode='wb') as f:
                hasKeys = False
                if export_type == "TAB":
                    csvwriter = csv.writer(f, delimiter='\t')
                else:
                    csvwriter = csv.writer(f)
                for message in messages['logMessages']:
                    if hasKeys == False:
                        csvwriter.writerow(message.keys())
                        hasKeys = True
                    csvwriter.writerow(message.values())
                    del message
            del messages
            return out_path
        else:
            return self._con.post(path=url,
                                  postdata=params)



"""
These functions are used to identify areas that meet a number of different criteria you specify. These criteria can be based
upon attribute queries (for example, parcels that are vacant) and spatial queries (for example, within 1 kilometer of a
river). The areas that are found can be selected from existing features (such as existing land parcels) or new features
can be created where all the requirements are met.

find_existing_locations searches for existing areas in a layer that meet a series of criteria.
derive_new_locations creates new areas from locations in your study area that meet a series of criteria.
find_similar_locations finds locations most similar to one or more reference locations based on criteria you specify.
choose_best_facilities choose the best locations for facilities by allocating locations that have demand for these
facilities in a way that satisfies a given goal.
create_viewshed creates areas that are visible based on locations you specify.
create_watersheds creates catchment areas based on locations you specify.
trace_downstream determines the flow paths in a downstream direction from the locations you specify
"""
import arcgis as _arcgis

def find_existing_locations(
        input_layers=[],
        expressions=[],
        output_name=None,
        context=None,
        gis=None):
    """
    The Find Existing Locations task selects features in the input layer that meet a query you specify.
    A query is made up of one or more expressions. There are two types of expressions: attribute and spatial.
    An example of an attribute expression is that a parcel must be vacant, which is an attribute of the Parcels layer
    (where STATUS = 'VACANT'). An example of a spatial expression is that the parcel must also be within a certain
    distance of a river (Parcels within a distance of 0.75 Miles from Rivers).

    Parameters
    ----------
    input_layers : Required list of strings
        A list of layers that will be used in the expressions parameter.
    expressions : Required string
        Specify a list of expressions. Please refer documentation at http://developers.arcgis.com for more information
        on creating expressions.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    result_layer : layer (FeatureCollection)
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_existing_locations(
        input_layers,
        expressions,
        output_name,
        context)


def derive_new_locations(
        input_layers=[],
        expressions=[],
        output_name=None,
        context=None,
        gis=None):
    """
    The Derive New Locations task derives new features from the input layers that meet a query you specify. A query is
    made up of one or more expressions. There are two types of expressions: attribute and spatial. An example of an
    attribute expression is that a parcel must be vacant, which is an attribute of the Parcels layer
    (where STATUS = 'VACANT'). An example of a spatial expression is that the parcel must also be within a certain
    distance of a river (Parcels within a distance of 0.75 Miles from Rivers).The Derive New Locations task is very
    similar to the Find Existing Locations task, the main difference is that the result of Derive New Locations can
    contain partial features.In both tasks, the attribute expression  where and the spatial relationships within and
    contains return the same result. This is because these relationships return entire features.When intersects or
    withinDistance is used, Derive New Locations creates new features in the result. For example, when intersecting a
    parcel feature and a flood zone area that partially overlap each other, Find Existing Locations will return the
    entire parcel whereas Derive New Locations will return just the portion of the parcel that is within the flood zone.

    Parameters
    ----------
    input_layers : Required list of strings
        A list of layers that will be used in the expressions parameter.
    expressions : Required string
        Specify a list of expressions. Please refer documentation at http://developers.arcgis.com for more information
        on expressions.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    result_layer : layer (FeatureCollection)
    """

    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.derive_new_locations(
        input_layers,
        expressions,
        output_name,
        context)


def find_similar_locations(
        input_layer,
        search_layer,
        analysis_fields=[],
        input_query=None,
        number_of_results=0,
        output_name=None,
        context=None,
        gis=None):
    """
    Finds the locations that are most similar to one or more reference locations based on criteria that you specify.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)

    search_layer : Required layer (see Feature Input in documentation)

    analysis_fields : Required list of strings

    input_query : Optional string

    number_of_results : Optional int

    output_name : Optional string

    context : Optional string


    Returns
    -------
    dict with the following keys:
       "similar_result_layer" : layer (FeatureCollection)
       "process_info" : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.find_similar_locations(
        input_layer,
        search_layer,
        analysis_fields,
        input_query,
        number_of_results,
        output_name,
        context)

"""
def choose_best_facilities():
    '
    Choose the best locations for facilities by allocating locations that have demand for these facilities in a way that
    satisfies a given goal.
    '
    pass #TODO
"""

def create_viewshed(
        input_layer,
        dem_resolution="Finest",
        maximum_distance=None,
        max_distance_units="Meters",
        observer_height=None,
        observer_height_units="Meters",
        target_height=None,
        target_height_units="Meters",
        generalize=True,
        output_name=None,
        context=None,
        gis=None):
    """
    Creates areas that are visible based on locations you specify.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)

    dem_resolution : Optional string

    maximum_distance : Optional float

    max_distance_units : Optional string

    observer_height : Optional float

    observer_height_units : Optional string

    target_height : Optional float

    target_height_units : Optional string

    generalize : Optional bool

    output_name : Optional string

    context : Optional string

    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


    Returns
    -------
    viewshed_layer : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.create_viewshed(
        input_layer,
        dem_resolution,
        maximum_distance,
        max_distance_units,
        observer_height,
        observer_height_units,
        target_height,
        target_height_units,
        generalize,
        output_name,
        context)


def create_watersheds(
        input_layer,
        search_distance=None,
        search_units="Meters",
        source_database="FINEST",
        generalize=True,
        output_name=None,
        context=None,
        gis=None):
    """
    Creates catchment areas based on locations you specify.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)

    search_distance : Optional float

    search_units : Optional string

    source_database : Optional string

    generalize : Optional bool

    output_name : Optional string

    context : Optional string

    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


    Returns
    -------
    dict with the following keys:
       "snap_pour_pts_layer" : layer (FeatureCollection)
       "watershed_layer" : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.create_watersheds(
        input_layer,
        search_distance,
        search_units,
        source_database,
        generalize,
        output_name,
        context)


def trace_downstream(
        input_layer,
        split_distance=None,
        split_units="Kilometers",
        max_distance=None,
        max_distance_units="Kilometers",
        bounding_polygon_layer=None,
        source_database=None,
        generalize=True,
        output_name=None,
        context=None,
        gis=None):
    """
    Determine the flow paths in a downstream direction from the locations you specify.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)

    split_distance : Optional float

    split_units : Optional string

    max_distance : Optional float

    max_distance_units : Optional string

    bounding_polygon_layer : Optional layer (see Feature Input in documentation)

    source_database : Optional string

    generalize : Optional bool

    output_name : Optional string

    context : Optional string

    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    trace_layer : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.trace_downstream(
        input_layer,
        split_distance,
        split_units,
        max_distance,
        max_distance_units,
        bounding_polygon_layer,
        source_database,
        generalize,
        output_name,
        context)



"""

These tools are used for the day-to-day management of geographic and tabular data.

copy_to_data_store copies data to your ArcGIS Data Store and creates a layer in your web GIS.
"""
import json as _json
import logging as _logging
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet
from arcgis.geoprocessing._support import _execute_gp_tool
from ._util import _id_generator, _feature_input, _set_context, _create_output_service

_log = _logging.getLogger(__name__)

_use_async = True

def overlay_data(input_layer, overlay_layer, overlay_type="intersect", output_name=None, gis=None):
    """
    Only available at ArcGIS Enterprise 10.6.1 and later.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       required FeatureLayer. The point, line or polygon features.
    ----------------  ---------------------------------------------------------------
    overlay_layer     required FeatureLayer. The features that will be overlaid with the input_layer features.
    ----------------  ---------------------------------------------------------------
    overlay_type      optional string. The type of overlay to be performed.
                      Values: intersect, erase

                      + intersect - Computes a geometric intersection of the input layers. Features or portions of features that overlap in both the inputLayer and overlayLayer layers will be written to the output layer. This is the default.
                      + erase - Only those features or portions of features in the overlay_layer that are not within the features in the input_layer layer are written to the output.
    ----------------  ---------------------------------------------------------------
    output_name       optional string. The task will create a feature service of the results. You define the name of the service.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS. The GIS object where the analysis will take place.
    ================  ===============================================================

    :returns: FeatureLayer
    """
    kwargs = locals()
    tool_name = "OverlayLayers"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
        "outputType" : "Input",
        'tolerance' : 0,
        'snapToInput' : 'false'
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Overlay_Layers_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Overlay Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "overlay_layer": (_FeatureSet, "overlayLayer"),
        "outputType" : (str, 'outputType'),
        "overlay_type" : (str, "overlayType"),
        "output_name": (str, "OutputName"),
        "context": (str, "context"),
        'tolerance' : (int, 'tolerance'),
        "output": (_FeatureSet, "output"),
        'snapToInput' : (str, 'snapToInput')
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

    return


def append_data(input_layer, append_layer, field_mapping=None, gis=None):
    """
    Only available at ArcGIS Enterprise 10.6.1 and later.

    The Append Data task appends tabular, point, line, or polygon data to an existing layer.
    The input layer must be a hosted feature layer. The tool will add the appended data as
    rows to the input layer. No new output layer is created.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       required FeatureLayer , The table, point, line or polygon features.
    ----------------  ---------------------------------------------------------------
    append_layer      required FeatureLayer. The table, point, line, or polygon features
                      to be appended to the input_layer. To append geometry, the
                      append_layer must have the same geometry type as the
                      input_layer. If the geometry types are not the same, the
                      append_layer geometry will be removed and all other matching
                      fields will be appended. The geometry of the input_layer will
                      always be maintained.
    ----------------  ---------------------------------------------------------------
    field_mapping     Defines how the fields in append_layer are appended to the
                      input_layer.

                      The following are set by default:

                        - All append_layer fields that match input_layer schema will be appended.
                        - Fields that exist in the input_layer and not in the append_layer will be appended with null values.
                        - Fields that exist in the append_layer and not in the input_layer will not be appended.

                      Optionally choose how input_layer fields will be appended from the following:

                      - AppendField - Matches the input_layer field with an append_layer field of a different name. Field types must match.
                      - Expression - Calculates values for the resulting field. Values are calculated using Arcade expressions. To assign null values, use 'null'.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS, the GIS on which this tool runs. If not
                      specified, the active GIS is used.
    ================  ===============================================================

    :returns: boolean

    """
    kwargs = locals()
    tool_name = "AppendData"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json"
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "append_layer": (_FeatureSet, "appendLayer"),
        "field_mapping" : (str, "fieldMapping"),
        "context": (str, "context")
    }
    return_values = [
    ]
    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return True
    except:
        raise

    return False


def calculate_fields(input_layer,
                     field_name,
                     data_type,
                     expression,
                     track_aware=False,
                     track_fields=None,
                     output_name=None,
                     gis=None
                     ):
    """
    The Calculate Field task works with a layer to create and populate a
    new field. The output is a new feature layer, that is the same as the
    input features, with the additional field added.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       required service , The table, point, line or polygon features
                      containing potential incidents.
    ----------------  ---------------------------------------------------------------
    field_name        required string, A string representing the name of the new
                      field. If the name already exists in the dataset, then a
                      numeric value will be appended to the field name.
    ----------------  ---------------------------------------------------------------
    data_type         required string, the type for the new field.
                      Values: Date |Double | Integer | String
    ----------------  ---------------------------------------------------------------
    expression        required string, An Arcade expression used to calculate the new
                      field values. You can use any of the Date, Logical,
                      Mathematical or Text function available with Arcade.
    ----------------  ---------------------------------------------------------------
    track_aware       optional boolean, Boolean value denoting if the expression is
                      track aware.
                      Default: False
    ----------------  ---------------------------------------------------------------
    track_fields      optional string, The fields used to identify distinct tracks.
                      There can be multiple track_fields. track_fields are only
                      required when track_aware is true.
    ----------------  ---------------------------------------------------------------
    output_name       optional string, The task will create a feature service of the
                      results. You define the name of the service.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS, the GIS on which this tool runs. If not
                      specified, the active GIS is used.
    ================  ===============================================================

    :returns:
       Feature Layer
    """
    kwargs = locals()
    tool_name = "CalculateField"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json"
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Calculate_Fields_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Calculate Fields')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "field_name" : (str, "fieldName"),
        "data_type" : (str, "dataType"),
        "expression" : (str, "expression"),
        "track_aware" : (bool, "trackAware"),
        "track_fields" : (str, "trackFields"),
        "output_name": (str, "outputName"),
        "output": (_FeatureSet, "output"),
        "context": (str, "context")
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

    return

def copy_to_data_store(
    input_layer,
    output_name = None,
    gis = None):
    """

    Copies an input feature layer or table to an ArcGIS Data Store and creates a layer in your web GIS.

    For example

    * Copy a collection of .csv files in a big data file share to the spatiotemporal data store for visualization.

    * Copy the features in the current map extent that are stored in the spatiotemporal data store to the relational data store.

    This tool will take an input layer and copy it to a data store. Data will be copied to the ArcGIS Data Store and will be stored in your relational or spatiotemporal data store.

    For example, you could copy features that are stored in a big data file share to a relational data store and specify that only features within the current map extent will be copied. This would create a hosted feature service with only those features that were within the specified map extent.

   Parameters:

   input_layer: Input Layer (feature layer). Required parameter.

   output_name: Output Layer Name (str). Required parameter.

   gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   output - Output Layer as a feature layer collection item


    """
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Data Store Copy_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Copy To Data Store')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Layer"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Layer", "type": _FeatureSet},
    ]
    try:
        _execute_gp_tool(gis, "CopyToDataStore", params, param_db, return_values, _use_async, url, True)
        return output_service
    except:
        output_service.delete()
        raise

copy_to_data_store.__annotations__ = {
    'output_name': str}








""" Defines the Worker class.
"""

from ..workforce import _store
from .exceptions import ValidationError
from .feature_model import FeatureModel
from ._store import *
from ._schemas import WorkerSchema


class Worker(FeatureModel):
    """
    Represents a worker in a Workforce Project

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    project                Required :class:`~arcgis.apps.workforce.Project`. The project that
                           the worker belongs to.
    ------------------     --------------------------------------------------------------------
    feature                Optional :class:`~arcgis.features.Feature`. The feature representing
                           the worker. Mostly intended for
                           internal usage. If supplied, other parameters are ignored.
    ------------------     --------------------------------------------------------------------
    geometry               Optional :class:`Dict`. The geometry of the worker.
    ------------------     --------------------------------------------------------------------
    contact_number         Optional :class:`String`. The contact number of the worker.
    ------------------     --------------------------------------------------------------------
    name                   Optional :class:`String`. The name of the worker.
    ------------------     --------------------------------------------------------------------
    notes                  Optional :class:`String`. The notes about the worker.
    ------------------     --------------------------------------------------------------------
    status                 Optional :class:`String`. The status of the worker.

                           `not_working`, `working`, `on_break`
    ------------------     --------------------------------------------------------------------
    title                  Optional :class:`String`. The title of the worker.
    ------------------     --------------------------------------------------------------------
    user_id                Optional :class:`String`. The user id of the worker
    ==================     ====================================================================

    """

    def __init__(self, project, feature=None, geometry=None, contact_number=None,
                 name=None, notes=None, status=None, title=None, user_id=None):
        super().__init__(project, project.workers_layer, feature)
        self._schema = WorkerSchema(project.workers_layer)
        if not feature:
            self.geometry = geometry
            self.contact_number = contact_number
            self.name = name
            self.notes = notes
            self.status = status
            self.title = title
            self.user_id = user_id

    def __str__(self):
        return "{} ({})".format(self.name, self.user_id)

    def __repr__(self):
        return "<Worker {}>".format(self.object_id)

    def update(self, geometry=None, contact_number=None,
                 name=None, notes=None, status=None, title=None, user_id=None):
        """
            Updates the worker on the server

            ==================     ====================================================================
            **Argument**           **Description**
            ------------------     --------------------------------------------------------------------
            geometry               Optional :class:`Dict`. The geometry of the worker.
            ------------------     --------------------------------------------------------------------
            contact_number         Optional :class:`String`. The contact number of the worker.
            ------------------     --------------------------------------------------------------------
            name                   Optional :class:`String`. The name of the worker.
            ------------------     --------------------------------------------------------------------
            notes                  Optional :class:`String`. The notes about the worker.
            ------------------     --------------------------------------------------------------------
            status                 Optional :class:`String`. The status of the worker.

                                   `not_working`, `working`, `on_break`
            ------------------     --------------------------------------------------------------------
            title                  Optional :class:`String`. The title of the worker.
            ------------------     --------------------------------------------------------------------
            user_id                Optional :class:`String`. The user id of the worker
            ==================     ====================================================================

        """
        update_worker(self.project, self, geometry, contact_number, name, notes, status, title, user_id)

    def delete(self):
        """Deletes the worker from the server"""
        delete_workers(self.project, [self])

    @property
    def name(self):
        """Gets/Sets the name of the worker"""
        return self._feature.attributes.get(self._schema.name)

    @name.setter
    def name(self, value):
        self._feature.attributes[self._schema.name] = value

    @property
    def contact_number(self):
        """Gets/Sets the contact number of the worker"""
        return self._feature.attributes.get(self._schema.contact_number)

    @contact_number.setter
    def contact_number(self, value):
        self._feature.attributes[self._schema.contact_number] = value

    @property
    def title(self):
        """Gets/Sets the title of the worker"""
        return self._feature.attributes.get(self._schema.title)

    @title.setter
    def title(self, value):
        self._feature.attributes[self._schema.title] = value

    @property
    def notes(self):
        """Gets/Sets the notes of the worker"""
        return self._feature.attributes.get(self._schema.notes)

    @notes.setter
    def notes(self, value):
        self._feature.attributes[self._schema.notes] = value

    @property
    def user_id(self):
        """Gets/Sets the user id of the worker"""
        return self._feature.attributes.get(self._schema.user_id)

    @user_id.setter
    def user_id(self, value):
        self._feature.attributes[self._schema.user_id] = value

    @property
    def status(self):
        """
            Gets/Sets the :class:`String` status of the worker

            `not_working`, `working`, `on_break`
        """
        lut = {
            0: "not_working",
            1: "working",
            2: "on_break",
        }
        if self._feature.attributes[self._schema.status] is not None:
            return lut[self._feature.attributes[self._schema.status]]
        else:
            return None

    @status.setter
    def status(self, value):
        if (isinstance(value, int) and value >= 0 and value <= 2) or value is None:
            self._feature.attributes[self._schema.status] = value
        elif isinstance(value, str):
            reduced_str = value.lower().replace(" ", "").replace("_", "")
            if reduced_str == "notworking":
                self._feature.attributes[self._schema.status] = 0
            elif reduced_str == "working":
               self._feature.attributes[self._schema.status] = 1
            elif reduced_str == "onbreak":
               self._feature.attributes[self._schema.status] = 2
            else:
               raise ValidationError("Invalid status", self)
        else:
            raise ValidationError("Invalid status", self)

    @FeatureModel.geometry.setter
    def geometry(self, value):
        self._feature.geometry = value

    def _validate(self, **kwargs):
        """
        """
        errors = super()._validate(**kwargs)
        errors += self._validate_name()
        errors += self._validate_status()
        errors += self._validate_user_id()
        return errors

    def _validate_for_add(self, **kwargs):
        errors = super()._validate_for_add(**kwargs)
        errors += self._validate_user_id_on_server()
        return errors

    def _validate_for_update(self, **kwargs):
        errors = super()._validate_for_update(**kwargs)
        errors += self._validate_user_id_on_server()
        return errors

    def _validate_for_remove(self, **kwargs):
        errors = super()._validate_for_remove(**kwargs)
        assignments = _store.query_assignments(self.project, "{} = {}".format(self.project._assignment_schema.worker_id, self.object_id))
        if assignments:
            errors.append(ValidationError("Cannot remove a Worker that has assignments", self))
        return errors

    def _validate_name(self):
        errors = []
        if not self.name or self.name.isspace():
            errors.append(ValidationError("Worker cannot have an empty name", self))
        return errors

    def _validate_status(self):
        errors = []
        if self.status is None:
            errors.append(ValidationError("Worker must have a status", self))
        return errors

    def _validate_user_id(self):
        errors = []
        if not self.user_id or self.user_id.isspace():
            errors.append(ValidationError("Worker cannot have an empty user_id", self))
        return errors

    def _validate_user_id_on_server(self):
        errors = []
        user = self.project.gis.users.get(self.user_id)
        if user is None:
            message = "The Worker user_id must match an accessible named user id"
            errors.append(ValidationError(message, self))

        workers = [w for w in self.project._cached_workers.values() if w.user_id == self.user_id]
        duplicate_workers = [w for w in workers if w.object_id != self.object_id]
        if duplicate_workers:
            message = "There cannot be multiple Workers with the same user_id"
            errors.append(ValidationError(message, self))
        return errors



""" Defines store functions for working with Projects.
"""

from arcgis.gis import Item

from ... import workforce


def get_project(project_id, gis):
    """ Loads and returns a workforce project.
        :param gis: An authenticated arcigs.gis.GIS object.
        :param project_id: The project's id.
        :returns: workforce.Project
    """
    item = Item(gis, project_id)
    return workforce.Project(item)



########################################################################
class CreditManager(object):
    """
    Manages an AGOL Site's Credits for users and sites
    """
    _gis = None
    _con = None
    _portal = None
    #----------------------------------------------------------------------
    def __init__(self, gis):
        """Constructor"""
        self._gis = gis
        self._portal = gis._portal
        self._con = self._portal.con
    #----------------------------------------------------------------------
    @property
    def credits(self):
        """returns the current number of credits on the GIS"""
        try:
            return self._gis.properties.availableCredits
        except:
            return 0
    #----------------------------------------------------------------------
    @property
    def is_enabled(self):
        """
        boolean that show is credit credit assignment
        """
        return self._gis.properties.creditAssignments == 'enabled'
    #----------------------------------------------------------------------
    def enable(self):
        """
        enables credit allocation on AGOL
        """
        return self._gis.update_properties(
            {"creditAssignments" : 'enabled'})
    #----------------------------------------------------------------------
    def disable(self):
        """
        disables credit allocation on AGOL
        """
        return self._gis.update_properties(
            {"creditAssignments" : 'disabled'})
    #----------------------------------------------------------------------
    @property
    def default_limit(self):
        """
        Gets/Sets the default credit allocation for AGOL
        """
        return self._gis.properties.defaultUserCreditAssignment
    #----------------------------------------------------------------------
    @default_limit.setter
    def default_limit(self, value):
        """
        Gets/Sets the default credit allocation for AGOL
        """
        params = {"defaultUserCreditAssignment" : value}
        self._gis.update_properties(params)
    #----------------------------------------------------------------------
    def allocate(self, username, credits=None):
        """
        Allows organization administrators to allocate credits for
        organizational users in ArcGIS Online

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        username                        Required string.The name of the user to assign credits to.
        ---------------------------     --------------------------------------------------------------------
        credits                         Optional float. The number of credits to assign to a user. If None
                                        is provided, it sets user to unlimited credits.
        ===========================     ====================================================================

        :returns: boolean

        """
        if credits:
            params = {
                "f" : "json",
                "userAssignments" : [{"username" : username, "credits" : credits}]
            }
            path = "portals/self/assignUserCredits"
            res =  self._con.post(path, params)
            if 'success' in res:
                return res['success']
            return res
        else:
            return self.deallocate(username=username)
    #----------------------------------------------------------------------
    def deallocate(self, username):
        """
        Allows organization administrators to remove credit allocation for
        organizational users in ArcGIS Online

        ===========================     ====================================================================
        **Argument**                    **Description**
        ---------------------------     --------------------------------------------------------------------
        username                        Required string.The name of the user to remove all credit from.
        ===========================     ====================================================================

        :returns: boolean

        """
        params = {"usernames" : [username],
                  "f" : 'json'}
        path = "portals/self/unassignUserCredits"
        res = self._con.post(path, params)
        if 'success' in res:
            return res['success']
        return res
    #----------------------------------------------------------------------
    def credit_usage(self, start_time=None, end_time=None):
        """
        returns the total credit consumption for a given time period.

        ===================   ===============================================
        **arguements**        **description**
        -------------------   -----------------------------------------------
        start_time            datetime.datetime object. This is the date to
                              start at.
        -------------------   -----------------------------------------------
        end_time              datetime.datetime object. This is the stop time
                              to look for credit consumption. It needs to be
                              at least 1 day previous than then start_time.
        ===================   ===============================================

        returns: dictionary
        """
        import datetime
        if isinstance(start_time, datetime.datetime):
            start_time =int(start_time.timestamp() * 1000)
        else:
            start_time = int(datetime.datetime.now().timestamp() * 1000)
        if isinstance(end_time, datetime.datetime):
            end_time = int(end_time.timestamp() * 1000)
        else:
            end_time = int((datetime.datetime.now() - datetime.timedelta(days=5)).timestamp() * 1000)
        path = "portals/self/usage"
        params = {
        'f' : 'json',
        'startTime' : end_time,
        'endTime' : start_time,
        'period' : '1d',
        'groupby' : 'stype,etype',
        'vars' : 'credits,num'
        }
        data = self._con.get(path, params)
        res = {}
        for d in data['data']:
            if d['stype'] in res:
                res[d['stype']] += sum([float(a[1]) for a in d['credits']])
            else:
                res[d['stype']] = sum([float(a[1]) for a in d['credits']])
        return res



from __future__ import absolute_import
from __future__ import print_function
import json
from arcgis.geometry import Polygon, Polyline, Point, MultiPoint
########################################################################
class StatisticFilter(object):
    """
    The definitions for one or more field-based statistics to be calculated
    """
    _json = None
    _array = []

    #----------------------------------------------------------------------
    def __init__(self):
        """Constructor"""
        pass
    #----------------------------------------------------------------------
    def add(self, statisticType, onStatisticField, outStatisticFieldName=None):
        """
        Adds the statistics group to the filter.

        outStatistics - is supported on only those layers/tables that
          indicate supportsStatistics is true.
        outStatisticFieldName is empty or missing, the map server assigns a
          field name to the returned statistic field. A valid field name
          can only contain alphanumeric characters and an underscore.
        outStatisticFieldName is a reserved keyword of the underlying DBMS,
          the operation can fail. Try specifying an alternative
          outStatisticFieldName. When using outStatistics, the only other
          parameters that can be used are groupByFieldsForStatistics,
          orderByFields, time, and where.
        """
        val = {
            "statisticType" : statisticType,
            "onStatisticField" : onStatisticField,
            "outStatisticFieldName" : outStatisticFieldName
        }
        if outStatisticFieldName is None:
            del val['outStatisticFieldName']
        self._array.append(val)
    #----------------------------------------------------------------------
    def remove(self, index):
        """removes the filter by index"""
        self._array.remove(index)
    #----------------------------------------------------------------------
    def clear(self):
        """removes all the filters"""
        self._array = []
    #----------------------------------------------------------------------
    @property
    def filter(self):
        """ returns the key/value pair of a geometry filter """
        return self._array
########################################################################
class LayerDefinitionFilter(object):
    """
       Allows you to filter the features of individual layers in the
       query by specifying definition expressions for those layers. A
       definition expression for a layer that is published with the
       service will always be honored.
    """
    _ids = []
    _filterTemplate = {"layerId" : "", "where" : "", "outFields" : "*"}
    _filter = []
    #----------------------------------------------------------------------
    def __init__(self):
        """Constructor"""
        pass
    #----------------------------------------------------------------------
    def addFilter(self, layer_id, where=None, outFields="*"):
        """ adds a layer definition filter """
        import copy
        f = copy.deepcopy(self._filterTemplate)
        f['layerId'] = layer_id
        f['outFields'] = outFields
        if where is not None:
            f['where'] = where
        if f not in self._filter:
            self._filter.append(f)
    #----------------------------------------------------------------------
    def removeFilter(self, filter_index):
        """ removes a layer filter based on position in filter list """
        f = self._filter[filter_index]
        self._filter.remove(f)
    #----------------------------------------------------------------------
    def removeAll(self):
        """ removes all items from the filter """
        self._filter = []
    #----------------------------------------------------------------------
    @property
    def filter(self):
        """ returns the filter object as a list of layer defs """
        return self._filter
########################################################################
class GeometryFilter(object):
    """ creates a geometry filter for queries
        Inputs:
           geomObject - a common.Geometry or arcpy.Geometry object
           spatialFilter - The spatial relationship to be applied on the
                           input geometry while performing the query. The
                           supported spatial relationships include
                           intersects, contains, envelope intersects,
                           within, etc. The default spatial relationship
                           is intersects (esriSpatialRelIntersects).
           bufferDistance - if filter type esriSpatialRelWithin is selected
                            and the service supports that select type, then
                            the geometry will be buffered at a given.
                            Can be of type integer or float.
           units - the value the distance units represents. Valid values
                   are: "esriSRUnit_Meter", "esriSRUnit_StatuteMile",
                        "esriSRUnit_Foot", "esriSRUnit_Kilometer",
                         "esriSRUnit_NauticalMile", and
                         "esriSRUnit_USNauticalMile"
       Raises:
          AttributeError for invalid inputs
    """
    _allowedFilters = ["esriSpatialRelIntersects",
                       "esriSpatialRelContains",
                       "esriSpatialRelCrosses",
                       "esriSpatialRelEnvelopeIntersects",
                       "esriSpatialRelIndexIntersects",
                       "esriSpatialRelOverlaps",
                       "esriSpatialRelTouches",
                       "esriSpatialRelWithin"]
    _geomObject = None
    _spatialAction = None
    _geomType = None
    _spatialReference = None
    _buffer = None
    _units = None
    _allowed_units = ["esriSRUnit_Meter", "esriSRUnit_StatuteMile",
                      "esriSRUnit_Foot", "esriSRUnit_Kilometer",
                      "esriSRUnit_NauticalMile", "esriSRUnit_USNauticalMile"]
    #----------------------------------------------------------------------
    def __init__(self,
                 geomObject,
                 spatialFilter="esriSpatialRelIntersects",
                 bufferDistance=None,
                 units="esriSRUnit_Meter"
                 ):
        """Constructor"""
        self.geometry = geomObject
        if spatialFilter in self._allowedFilters:
            self._spatialAction = spatialFilter
            self._spatialReference = self.geometry.spatialReference
        else:
            raise AttributeError("geomObject must be a geometry object and "+ \
                                 "spatialFilter must be of value: " + \
                                 "%s" % ", ".join(self._allowedFilters))
        if not bufferDistance is None and \
           isinstance(bufferDistance, (int, float)) and \
           not units is None and \
           units.lower() in [f.lower() for f in self._allowed_units]:
            self._buffer = bufferDistance
            self._units = units

    #----------------------------------------------------------------------
    @property
    def spatialRelation(self):
        """ gets the filter type """
        return self._spatialAction
    #----------------------------------------------------------------------
    @spatialRelation.setter
    def spatialRelation(self, value):
        if value.lower() in \
           [x.lower() for x in self._allowedFilters]:
            self._spatialAction = value
        else:
            raise AttributeError("spatialRelation must be values of " + \
                                 "%s" % ", ".join(self._allowedFilters))
    #----------------------------------------------------------------------
    @property
    def geometryType(self):
        """ returns the geometry type """
        return self._geomObject.type
    #----------------------------------------------------------------------
    @property
    def geometry(self):
        """ gets the geometry object used by the filter """
        return self._geomObject
    #----------------------------------------------------------------------
    @geometry.setter
    def geometry(self, geometry):
        """ sets the geometry value """

        if isinstance(geometry, (Polygon, Point, Polyline, MultiPoint)):
            self._geomObject = geometry
            self._geomType = geometry.type
        else:
            raise AttributeError("geometry must be a common.Geometry type.")
    #----------------------------------------------------------------------
    @property
    def filter(self):
        """ returns the key/value pair of a geometry filter """

        val = {"geometryType":self.geometryType,
                "geometry": json.dumps(self._geomObject.asDictionary),
                "spatialRel": self.spatialRelation,
                "inSR" : self._geomObject.spatialReference['wkid']}
        if self._buffer is not None and \
           self._units is not None:
            val['buffer'] = self._buffer
            val['units'] = self._units
        return val
########################################################################
class TimeFilter(object):
    """ Implements the time filter """
    _startTime = None
    _endTime = None
    #----------------------------------------------------------------------
    def __init__(self, start_time, time_zone="UTC", end_time=None):
        """Constructor"""
        self._startTime = start_time
        self._endTime = end_time
        self._tz = time_zone
    #----------------------------------------------------------------------
    @property
    def filter(self):
        if not self._endTime is None:
            val = "%s, %s" % (self._startTime, self._endTime)
            return val
        else:
            return "%s" % self._startTime


"""
Functions to filter query results by a spatial relationship with another geometry
Used when querying feature layers and imagery layers.
"""

from . import Geometry

# esriSpatialRelIntersects | esriSpatialRelContains | esriSpatialRelCrosses | esriSpatialRelEnvelopeIntersects | \
# esriSpatialRelIndexIntersects | esriSpatialRelOverlaps | esriSpatialRelTouches | esriSpatialRelWithin

def _filter(geometry, sr, rel):
    if not isinstance(geometry, Geometry):
        geometry = Geometry(geometry)

    filter = {
        'geometry': geometry,
        'geometryType': 'esriGeometry' + geometry.type,
        'spatialRel': rel,
    }

    if sr is None:
        if 'spatialReference' in geometry:
            sr = geometry['spatialReference']

    if sr is not None:
        filter['inSR'] = sr

    return filter

def intersects(geometry, sr=None):
    """filters results whose geometry intersects with the specified geometry"""
    return _filter(geometry, sr, 'esriSpatialRelIntersects')

def contains(geometry, sr=None):
    """Returns a feature if its shape is wholly contained within the search geometry.
    Valid for all shape type combinations."""
    return _filter(geometry, sr, 'esriSpatialRelContains')

def crosses(geometry, sr=None):
    """Returns a feature if the intersection of the interiors of the two shapes is not empty and has a lower dimension
     than the maximum dimension of the two shapes. Two lines that share an endpoint in common do not cross.
     Valid for Line/Line, Line/Area, Multi-point/Area, and Multi-point/Line shape type combinations."""
    return _filter(geometry, sr, 'esriSpatialRelCrosses')

def envelope_intersects(geometry, sr=None):
    """Returns features if the envelope of the two shapes intersects. """
    return _filter(geometry, sr, 'esriSpatialRelEnvelopeIntersects')

def index_intersects(geometry, sr=None):
    """Returns a feature if the envelope of the query geometry intersects the index entry for the target geometry."""
    return _filter(geometry, sr, 'esriSpatialRelIndexIntersects')

def overlaps(geometry, sr=None):
    """Returns a feature if the intersection of the two shapes results in an object of the same dimension, but different
     from both of the shapes. Applies to Area/Area, Line/Line, and Multi-point/Multi-point shape type combinations."""
    return _filter(geometry, sr, 'esriSpatialRelOverlaps')

def touches(geometry, sr=None):
    """Returns a feature if the two shapes share a common boundary. However, the intersection of the interiors of the
    two shapes must be empty. In the Point/Line case, the point may touch an endpoint only of the line. Applies to all
    combinations except Point/Point."""
    return _filter(geometry, sr, 'esriSpatialRelTouches')

def within(geometry, sr=None):
    """Returns a feature if its shape wholly contains the search geometry. Valid for all shape type combinations."""
    return _filter(geometry, sr, 'esriSpatialRelWithin')



""" Defines the Project object.
"""
import arcgis
from arcgis.features import FeatureLayer
from arcgis.gis import Group
from arcgis._impl.common._utils import _lazy_property
import json

from ._schemas import *
from .managers import *


class Project:
    """
    A Workforce Project

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    item                   Required :class:`~arcgis.gis.Item`. The item that
                           the contains the project.
    ==================     ====================================================================

    .. code-block:: python

        # Get a Project and search the assignments and workers.

        import arcgis
        gis = arcgis.gis.GIS("https://arcgis.com", "<username>", "<password>")
        item = gis.content.get("<item-id>")
        project = arcgis.apps.workforce.Project(item)
        assignments = project.assignments.search()
        workers = project.workers.search()


    """

    def __init__(self, item):
        """
            :param item: The project's arcigs.gis.Item
        """
        self.gis = item._gis
        self._item = item
        self._item_data = item.get_data()
        self._assignment_schema = AssignmentSchema(self.assignments_layer)
        self._track_schema = TrackSchema(self.tracks_layer)
        self._worker_schema = WorkerSchema(self.workers_layer)
        self._dispatcher_schema = DispatcherSchema(self.dispatchers_layer)
        self._update_cached_objects()

    def _update_cached_objects(self):
        """
            Caches the types, workers, and dispatchers for quicker assignment creation when querying
            Should be called when querying assignments
        """
        self._cached_assignment_types = {a.code: a for a in self.assignment_types.search()}
        self._cached_workers = {w.id: w for w in self.workers.search()}
        self._cached_dispatchers = {d.id: d for d in self.dispatchers.search()}

    def __str__(self):
        return self.title

    def __repr__(self):
        return "<Project {}>".format(self.id)

    def delete(self):
        """
            Deletes the project, group, folder, layers, and webmaps.
            Assumes the currently signed in user owns the project or is an admin.
        """
        owner = self._item.owner
        self.assignments_item.protect(False)
        self.assignments_item.delete()
        self.tracks_item.protect(False)
        self.tracks_item.delete()
        self.workers_item.protect(False)
        self.workers_item.delete()
        self.dispatchers_item.protect(False)
        self.dispatchers_item.delete()
        self.dispatcher_webmap.item.protect(False)
        self.dispatcher_webmap.item.delete()
        self.worker_webmap.item.protect(False)
        self.worker_webmap.item.delete()
        self._item.protect(False)
        self._item.delete()
        self.group.delete()
        for folder in self.gis.users.get(owner).folders:
            if folder['id'] == self._item_data['folderId']:
                self.gis.content.delete_folder(folder['title'], owner=owner)

    def _update_data(self):
        self._item.update({
            "text": json.dumps(self._item_data)
        })

    def update(self, summary=None):
        """
        Updates the project on the server

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        summary                  Optional :class:`String`. The summary of the project.
        ==================     ====================================================================
        """
        item_properties = {}
        if summary:
            item_properties['snippet'] = summary
        self._item.update(item_properties)

    @property
    def _tracking_enabled(self):
        return self._item_data["tracks"]["enabled"]

    @_tracking_enabled.setter
    def _tracking_enabled(self, value):
        self._item_data["tracks"]["enabled"] = value
        self._update_data()

    @property
    def _tracking_interval(self):
        return self._item_data["tracks"]["updateInterval"]

    @_tracking_interval.setter
    def _tracking_interval(self, value):
        self._item_data["tracks"]["updateInterval"] = value
        self._update_data()

    @property
    def id(self):
        """The item id of the project"""
        return self._item['id']

    @property
    def title(self):
        """Gets the title of the project"""
        return self._item['title']

    @property
    def summary(self):
        """The title of the project"""
        return self._item['snippet']

    @summary.setter
    def summary(self, value):
        self._item['snippet'] = value

    @property
    def owner_user_id(self):
        """The user id of the project owner."""
        return self._item['owner']

    @property
    def version(self):
        """The version of the project"""
        return self._item_data['version']

    @_lazy_property
    def assignments_item(self):
        """The assignments :class:`~arcgis.gis.Item`"""
        return self.gis.content.get(self._item_data['assignments']['serviceItemId'])

    @property
    def assignments_layer_url(self):
        """The assignments feature layer url"""
        return self._item_data['assignments']['url']

    @_lazy_property
    def dispatchers_item(self):
        """The dispatchers :class:`~arcgis.gis.Item`"""
        return self.gis.content.get(self._item_data['dispatchers']['serviceItemId'])

    @property
    def dispatchers_layer_url(self):
        """The dispatchers feature layer url"""
        return self._item_data['dispatchers']['url']

    @_lazy_property
    def tracks_item(self):
        """The tracks :class:`~arcgis.gis.Item`"""
        return self.gis.content.get(self._item_data['tracks']['serviceItemId'])

    @property
    def tracks_layer_url(self):
        """The tracks feature layer url"""
        return self._item_data['tracks']['url']

    @property
    def workers_layer_url(self):
        """The workers feature layer url"""
        return self._item_data['workers']['url']

    @property
    def dispatcher_web_map_id(self):
        """The dispatcher webmap item id"""
        return self._item_data['dispatcherWebMapId']

    @property
    def worker_web_map_id(self):
        """The worker webmap item id"""
        return self._item_data['workerWebMapId']

    @property
    def group_id(self):
        """The group id that all project items are part of"""
        return self._item_data['groupId']

    @_lazy_property
    def owner(self):
        """The owner :class:`~arcgis.gis.User` of the project"""
        return self.gis.users.get(self.owner_user_id)

    @_lazy_property
    def assignments_layer(self):
        """The assignments :class:`~arcgis.features.FeatureLayer`"""
        return FeatureLayer(self.assignments_layer_url, self.gis)

    @_lazy_property
    def dispatchers_layer(self):
        """The dispatchers :class:`~arcgis.features.FeatureLayer`"""
        return FeatureLayer(self.dispatchers_layer_url, self.gis)

    @_lazy_property
    def tracks_layer(self):
        """The tracks :class:`~arcgis.features.FeatureLayer`"""
        return FeatureLayer(self.tracks_layer_url, self.gis)

    @_lazy_property
    def workers_item(self):
        """The workers :class:`~arcgis.gis.Item`"""
        return self.gis.content.get(self._item_data['workers']['serviceItemId'])

    @_lazy_property
    def workers_layer(self):
        """The workers :class:`~arcgis.features.FeatureLayer`"""
        return FeatureLayer(self.workers_layer_url, self.gis)

    @_lazy_property
    def dispatcher_webmap(self):
        """The dispatcher :class:`~arcgis.mapping.WebMap` for the project"""
        return arcgis.mapping.WebMap(self.gis.content.get(self.dispatcher_web_map_id))

    @_lazy_property
    def worker_webmap(self):
        """The worker :class:`~arcgis.mapping.WebMap` for the project"""
        return arcgis.mapping.WebMap(self.gis.content.get(self.worker_web_map_id))

    @_lazy_property
    def group(self):
        """The :class:`~arcgis.gis.Group` that the project resources are part of"""
        return Group(self.gis, self.group_id)

    @property
    def assignments(self):
        """The :class:`~arcgis.apps.workforce.managers.AssignmentManager` for the project"""
        return AssignmentManager(self)

    @property
    def workers(self):
        """The :class:`~arcgis.apps.workforce.managers.WorkerManager` for the project"""
        return WorkerManager(self)

    @property
    def dispatchers(self):
        """The :class:`~arcgis.apps.workforce.managers.DispatcherManager` for the project"""
        return DispatcherManager(self)

    @property
    def tracks(self):
        """The :class:`~arcgis.apps.workforce.managers.TrackManager` for the project"""
        return TrackManager(self)

    @property
    def assignment_types(self):
        """The :class:`~arcgis.apps.workforce.managers.AssignmentTypeManager` for the project"""
        return AssignmentTypeManager(self)



"""set of common utilities"""
import os
import sys
import time
import uuid
import zipfile
import datetime
from datetime import date
import tempfile
from contextlib import contextmanager
import six
import logging
import decimal
import functools
#----------------------------------------------------------------------
def create_uid():
    if six.PY2:
        return uuid.uuid4().get_hex()
    else:
        return uuid.uuid4().hex
#----------------------------------------------------------------------
def _date_handler(obj):
    import numpy
    if isinstance(obj, datetime.datetime) or isinstance(obj, date):
        return local_time_to_online(obj)
    elif isinstance(obj, (numpy.int32, numpy.int64)):
        return _date_handler(int(obj))
    elif isinstance(obj, decimal.Decimal):
        return float(obj)
    elif isinstance(obj, (numpy.float, numpy.float32, numpy.float64)):
        return float(obj)
    elif isinstance(obj, numpy.ndarray):
        return obj.tolist()
    else:
        return obj
#----------------------------------------------------------------------
def local_time_to_online(dt=None):
    """
       converts datetime object to a UTC timestamp for AGOL
       Inputs:
          dt - datetime object
       Output:
          Long value
    """
    if dt is None:
        dt = datetime.datetime.now()

    if isinstance(dt, datetime.datetime) and dt.tzinfo:
        dt = dt.astimezone()

    return int(time.mktime(dt.timetuple())  * 1000)
#----------------------------------------------------------------------
def online_time_to_string(value,timeFormat):
    """
       Converts a timestamp to date/time string
       Inputs:
          value - timestamp as long
          timeFormat - output date/time format
       Output:
          string
    """
    return datetime.datetime.fromtimestamp(value /1000).strftime(timeFormat)
#----------------------------------------------------------------------
def timestamp_to_datetime(timestamp):
    """
       Converts a timestamp to a datetime object
       Inputs:
          timestamp - timestamp value as Long
       output:
          datetime object
    """
    return datetime.datetime.fromtimestamp(timestamp /1000)

###########################################################################
class Error(Exception): pass
#--------------------------------------------------------------------------
@contextmanager
def _tempinput(data):
    temp = tempfile.NamedTemporaryFile(delete=False)
    temp.write((bytes(data, 'UTF-8')))
    temp.close()
    yield temp.name
    os.unlink(temp.name)
#--------------------------------------------------------------------------
def _lazy_property(fn):
    '''Decorator that makes a property lazy-evaluated.
    '''
    # http://stevenloria.com/lazy-evaluated-properties-in-python/
    attr_name = '_lazy_' + fn.__name__

    @property
    @functools.wraps(fn)
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)
    return _lazy_property
#--------------------------------------------------------------------------
def _is_shapefile(data):
    if zipfile.is_zipfile(data):
        zf = zipfile.ZipFile(data, 'r')
        namelist = zf.namelist()
        for name in namelist:
            if name.endswith('.shp') or name.endswith('.SHP'):
                return True
    return False
#--------------------------------------------------------------------------
def rot13(s):
    result = ""

    # Loop over characters.
    for v in s:
        # Convert to number with ord.
        c = ord(v)

        # Shift number back or forward.
        if c >= ord('a') and c <= ord('z'):
            if c > ord('m'):
                c -= 13
            else:
                c += 13
        elif c >= ord('A') and c <= ord('Z'):
            if c > ord('M'):
                c -= 13
            else:
                c += 13

        # Append to result.
        result += chr(c)

    # Return transformation.
    return result
#--------------------------------------------------------------------------
def zipws(path, outfile, keep=True):
    """
    compress the contents of a folder
    Parameters:
     :path: folder or folder contents to compress as a zip file
     :outfile: output file and location
     :keep: boolean - if true, the folder structure is kept, else just the
      files
    Output:
     path to a compressed zip file.
    """
    zipobj = zipfile.ZipFile(outfile, 'w', zipfile.ZIP_DEFLATED)
    path = os.path.normpath(path)
    for (dirpath, dirnames, filenames) in os.walk(path):
        for file in filenames:
            if not file.endswith('.lock') and\
               not file.endswith('.zip'):
                try:
                    if keep:
                        zipobj.write(
                            os.path.join(dirpath,
                                         file),
                            os.path.join(os.path.basename(path),
                                         os.path.join(dirpath,
                                                      file)[len(path)+len(os.sep):]))
                    else:
                        zipobj.write(os.path.join(dirpath, file),
                        os.path.join(dirpath[len(path):], file))
                except Exception:
                    pass
    zipobj.close()
    return outfile
#--------------------------------------------------------------------------
def _to_utf8(data):
    """ Converts strings and collections of strings from unicode to utf-8. """
    if isinstance(data, dict):
        return {_to_utf8(key): _to_utf8(value) \
                for key, value in data.items() if value is not None}
    elif isinstance(data, list):
        return [_to_utf8(element) for element in data]
    elif isinstance(data, str):
        return data
    elif isinstance(data, six.text_type):
        return data.encode('utf-8')
    elif isinstance(data, (float, six.integer_types)):
        return data
    else:
        return data
#--------------------------------------------------------------------------
class _DisableLogger():
    def __enter__(self):
        logging.disable(logging.CRITICAL)
    def __exit__(self, a, b, c):
        logging.disable(logging.NOTSET)

# --------------------------------------------------------------------------
def chunks(l, n):
    """yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]





class FeatureSchema:
    """ Describes the schema used to persist a FeatureModel, which may vary based on the particular
        backend in use (AGO vs Enterprise).
    """

    def __init__(self, feature_layer):
        """ Initializes a new FeatureSchema.
            :param feature_layer: The FeatureLayer used to persist the FeatureModel.
            :type feature_layer: arcgis.features.FeatureLayer
        """
        self._feature_layer = feature_layer

    @property
    def object_id(self):
        """Gets the object id field name"""
        return self._field_called('OBJECTID')['name']

    @property
    def global_id(self):
        """Gets the global id field name"""
        return self._field_called('GlobalID')['name']

    @property
    def creator(self):
        """Gets the creator field name"""
        return self._editor_tracking_field('creatorField')['name']

    @property
    def creation_date(self):
        """Gets the creation date field name"""
        return self._editor_tracking_field('creationDateField')['name']

    @property
    def editor(self):
        """Gets the editor field name"""
        return self._editor_tracking_field('editorField')['name']

    @property
    def edit_date(self):
        """Gets the edit date field name"""
        return self._editor_tracking_field('editDateField')['name']

    def _field_called(self, name):
        """ Finds a field with a name similar to name, using a case-insensitive comparison.
        """
        for field in self._feature_layer.properties.fields:
            if field['name'].lower() == name.lower():
                return field

    def _editor_tracking_field(self, name):
        """ Finds the editor tracking field identified by the name.  The name corresponds to one of
            the universal editor tracking field names used in the FeatureLayer metadata, including
            'creatorField', 'creationDateField', 'editorField', and 'editDateField'.
        """
        return self._field_called(self._feature_layer.properties.editFieldsInfo[name])


class AssignmentSchema(FeatureSchema):
    """ Describes the schema for an assignment :py:attr:`~arcgis.apps.workforce.Assignment.schema`
    This is useful for getting the field names, which may be different depending on if
    Workforce for ArcGIS is hosted on ArcGIS Online or on an Enterprise Deployment.
    """

    @property
    def assigned_date(self):
        """Gets the assigned date field name"""
        return self._field_called('assignedDate')['name']

    @property
    def assignment_read(self):
        """Gets the assignment read field name"""
        return self._field_called('assignmentRead')['name']

    @property
    def assignment_type(self):
        """Gets the assignment type field name"""
        return self._field_called('assignmentType')['name']

    @property
    def completed_date(self):
        """Gets the completed date field name"""
        return self._field_called('completedDate')['name']

    @property
    def declined_comment(self):
        """Gets the comment field name"""
        return self._field_called('declinedComment')['name']

    @property
    def declined_date(self):
        """Gets the declined date field name"""
        return self._field_called('declinedDate')['name']

    @property
    def description(self):
        """Gets the description field name"""
        return self._field_called('description')['name']

    @property
    def dispatcher_id(self):
        """Gets the dispatcher id field name"""
        return self._field_called('dispatcherId')['name']

    @property
    def due_date(self):
        """Gets the due date field name"""
        return self._field_called('dueDate')['name']

    @property
    def in_progress_date(self):
        """Gets the in progress date field name"""
        return self._field_called('inProgressDate')['name']

    @property
    def location(self):
        """Gets the location field name"""
        return self._field_called('location')['name']

    @property
    def notes(self):
        """Gets the notes field name"""
        return self._field_called('notes')['name']

    @property
    def paused_date(self):
        """Gets the paused date field name"""
        return self._field_called('pausedDate')['name']

    @property
    def priority(self):
        """Gets the priority field name"""
        return self._field_called('priority')['name']

    @property
    def status(self):
        """Gets the status field name"""
        return self._field_called('status')['name']

    @property
    def work_order_id(self):
        """Gets the work order id field name"""
        return self._field_called('workOrderId')['name']

    @property
    def worker_id(self):
        """Gets the worker id field name"""
        return self._field_called('workerId')['name']


class DispatcherSchema(FeatureSchema):
    """ Describes the schema for a dispatcher :py:attr:`~arcgis.apps.workforce.Dispatcher.schema`
    This is useful for getting the field names, which may be different depending on if
    Workforce for ArcGIS is hosted on ArcGIS Online or on an Enterprise Deployment.
    """

    @property
    def contact_number(self):
        """Gets the contact number field name"""
        return self._field_called('contactNumber')['name']

    @property
    def name(self):
        """Gets the name field name"""
        return self._field_called('name')['name']

    @property
    def user_id(self):
        """Gets the user id field name"""
        return self._field_called('userId')['name']


class TrackSchema(FeatureSchema):
    """ Describes the schema for a track :py:attr:`~arcgis.apps.workforce.Track.schema`
    This is useful for getting the field names, which may be different depending on if
    Workforce for ArcGIS is hosted on ArcGIS Online or on an Enterprise Deployment.
    """

    @property
    def accuracy(self):
        """Gets the accuracy field name"""
        return self._field_called('accuracy')['name']


class WorkerSchema(FeatureSchema):
    """ Describes the schema for a worker :py:attr:`~arcgis.apps.workforce.Worker.schema`
    This is useful for getting the field names, which may be different depending on if
    Workforce for ArcGIS is hosted on ArcGIS Online or on an Enterprise Deployment.
    """

    @property
    def contact_number(self):
        """Gets the contact number field name"""
        return self._field_called('contactNumber')['name']

    @property
    def name(self):
        """Gets the name field name"""
        return self._field_called('name')['name']

    @property
    def notes(self):
        """Gets the notes field name"""
        return self._field_called('notes')['name']

    @property
    def status(self):
        """Gets the status field name"""
        return self._field_called('status')['name']

    @property
    def title(self):
        """Gets the title field name"""
        return self._field_called('title')['name']

    @property
    def user_id(self):
        """Gets the user id field name"""
        return self._field_called('userId')['name']



"""
The arcgis.geometry module defines useful geometry types for working with geographic information and GIS functionality.
It provides functions which use geometric types as input and output as well as functions for easily converting
geometries between different representations.

Several functions accept geometries represented as dictionaries and the geometry objects in this module behave like them
as well as support the '.' (dot) notation providing attribute access.

Example:

.. code-block:: python

    pt = Point({"x" : -118.15, "y" : 33.80, "spatialReference" : {"wkid" : 4326}})
    print (pt.is_valid)
    print (pt.type) # POINT
    print (pt)
    print (pt.x, pt.y)

Example Polyline:

.. code-block:: python


    line = {
      "paths" : [[[-97.06138,32.837],[-97.06133,32.836],[-97.06124,32.834],[-97.06127,32.832]],
                 [[-97.06326,32.759],[-97.06298,32.755]]],
      "spatialReference" : {"wkid" : 4326}
    }
    polyline = Polyline(line)
    print(polyline)
    print(polyline.is_valid)

Example of invalid geometry:

.. code-block:: python

    line = {
      "paths" : [[[-97.06138],[-97.06133,32.836],[-97.06124,32.834],[-97.06127,32.832]],
                 [[-97.06326,32.759],[-97.06298,32.755]]],
      "spatialReference" : {"wkid" : 4326}
    }
    polyline = Polyline(line)
    print(polyline)
    print(polyline.is_valid) # False

The same pattern can be used repeated for Polygon, MultiPoint and SpatialReference.

You can create a Geometry even when you don't know the exact type. The Geometry constructor is able to figure out the
geometry type and returns the correct type as the example below demonstrates:

.. code-block:: python

    geom = Geometry({
      "rings" : [[[-97.06138,32.837],[-97.06133,32.836],[-97.06124,32.834],[-97.06127,32.832],
                  [-97.06138,32.837]],[[-97.06326,32.759],[-97.06298,32.755],[-97.06153,32.749],
                  [-97.06326,32.759]]],
      "spatialReference" : {"wkid" : 4326}
    })
    print (geom.type) # Polygon
    print (isinstance(geom, Polygon) # True

"""

from ._types import *
from .functions import *
from . import filters
